<?php

declare(strict_types=1);

namespace Drupal\track_usage;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Database\Connection;
use Drupal\Core\Database\Query\Delete;
use Drupal\Core\Entity\EntityFieldManagerInterface;
use Drupal\Core\Entity\EntityInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Entity\FieldableEntityInterface;
use Drupal\Core\Entity\TranslatableInterface;
use Drupal\language\ConfigurableLanguageInterface;
use Drupal\track_usage\Entity\TrackConfigInterface;
use Drupal\track_usage\Model\EntityRole;
use Drupal\track_usage\Model\Usage;
use Drupal\track_usage\Trait\BackwardsCompatibilityTrait;
use Drupal\track_usage\Trait\EntityUtilityTrait;
use Drupal\track_usage\Model\Operation;
use Drupal\track_usage\Trait\IdColumnTrait;

/**
 * Default implementation of the track_usage.recorder service.
 */
class Recorder implements RecorderInterface {

  use BackwardsCompatibilityTrait;
  use EntityUtilityTrait;
  use IdColumnTrait;

  public function __construct(
    protected readonly TrackerInterface $tracker,
    protected readonly Connection $db,
    protected readonly EntityTypeManagerInterface $entityTypeManager,
    protected readonly EntityFieldManagerInterface $entityFieldManager,
  ) {}

  /**
   * {@inheritdoc}
   */
  public function record(EntityInterface $entity, TrackConfigInterface $config): void {
    $this->recordRaw($entity->getEntityTypeId(), $entity->id(), $this->getRevisionId($entity), $config->id());
  }

  /**
   * {@inheritdoc}
   */
  public function recordRaw(string $entityTypeId, int|string $entityId, int|string $revisionId, string $configId): void {
    $entity = $this->loadEntity($entityTypeId, $entityId, intval($revisionId));
    if (!$entity) {
      return;
    }

    $config = $this->entityTypeManager->getStorage('track_usage_config')->load($configId);
    if (!$config) {
      return;
    }

    if ($config->isTraversable($entity)) {
      assert($entity instanceof FieldableEntityInterface);
      // Re-queue all source entities where this entity is involved.
      $this->requeueSourceEntities($entity, $config);
      return;
    }

    if (!$config->isSource($entity)) {
      return;
    }

    assert($entity instanceof FieldableEntityInterface);

    $entityTypeId = $entity->getEntityTypeId();
    $entityIdAsInteger = $this->getIntegerIdColumnValue($entity);
    $entityIdAsString = $this->getStringIdColumnValue($entity);

    foreach ($this->getTranslations($entity) as $translation) {
      foreach ($this->tracker->track($translation, $config) as $usage) {
        assert($usage instanceof Usage);
        $meta = json_encode((object) ['paths' => $usage->paths]);
        $this->db->merge(self::TABLE)->keys([
          'config' => $config->id(),
          'source_type' => $entityTypeId,
          'source_id_int' => $entityIdAsInteger,
          'source_id_string' => $entityIdAsString,
          'source_langcode' => $translation->language()->getId(),
          'source_revision' => $this->getRevisionId($translation),
          'target_type' => $usage->targetEntity->getEntityTypeId(),
          'target_id_int' => $this->getIntegerIdColumnValue($usage->targetEntity),
          'target_id_string' => $this->getStringIdColumnValue($usage->targetEntity),
        ])->fields(['meta' => $meta])->execute();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function cleanup(EntityInterface $entity, Operation $operation): void {
    match($operation) {
      Operation::Delete => $this->onEntityDelete($entity),
      Operation::DeleteRevision => $this->onEntityRevisionDelete($entity),
      Operation::DeleteTranslation => $this->onEntityTranslationDelete($entity),
    };

    // The deleted entity might be a traversable entity, part of usage paths.
    if ($entity instanceof FieldableEntityInterface) {
      $configs = $this->entityTypeManager->getStorage('track_usage_config')->loadMultiple();
      foreach ($configs as $config) {
        if ($config->isTraversable($entity)) {
          assert($entity instanceof FieldableEntityInterface);
          // Re-queue all source entities where this entity is involved.
          $this->requeueSourceEntities($entity, $config);
        }
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getTargetsForEntity(FieldableEntityInterface $entity, string $targetEntityTypeId, TrackConfigInterface $config): iterable {
    if (!$config->isSource($entity) && !$config->isTraversable($entity)) {
      return;
    }

    $ids = $this->db->select(self::TABLE)
      ->distinct()
      ->fields(self::TABLE, ['target_id_string'])
      ->condition('target_type', $targetEntityTypeId)
      ->where("JSON_CONTAINS(meta, JSON_QUOTE(:entity), '\$.paths')", [
        ':entity' => $this->getKey($entity),
      ])
      ->execute()
      ->fetchCol();

    foreach ($ids as $id) {
      yield $id;
    }
  }

  /**
   * {@inheritdoc}
   */
  public function getPathsToTarget(EntityInterface $target, TrackConfigInterface $config): iterable {
    $yielded = [];

    $query = $this->db->select(self::TABLE)
      ->condition('config', $config->id())
      ->condition('target_type', $target->getEntityTypeId())
      ->condition($this->getIdColumnName($target, EntityRole::Target), $this->getIdColumnValueByEntity($target));
    $query->addExpression("JSON_EXTRACT(meta, '$.paths')", 'paths');

    foreach ($query->execute()->fetchAll($this->fetchMode(\PDO::FETCH_ASSOC)) as $row) {
      $paths = Json::decode($row['paths']);
      foreach ($paths as $path) {
        // Convert from key to array. E.g., 'media:1:2' -> ['media', '1', '2'].
        $path = array_map(fn(string $key): array => explode(':', $key, 3), $path);

        if (!in_array($path, $yielded, TRUE)) {
          $yielded[] = $path;
          yield $path;
        }
      }
    }
  }

  /**
   * Reacts after an entity has been deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity being deleted.
   */
  protected function onEntityDelete(EntityInterface $entity): void {
    // A track usage config was deleted. Remove all its records.
    if ($entity instanceof TrackConfigInterface) {
      $this->db->delete(self::TABLE)->condition('config', $entity->id())->execute();
      return;
    }
    // A language has been deleted.
    elseif ($entity instanceof ConfigurableLanguageInterface) {
      $this->db->delete(self::TABLE)->condition('source_langcode', $entity->id())->execute();
      return;
    }

    $entityTypeId = $entity->getEntityTypeId();
    $entityId = $this->getIdColumnValueByEntity($entity);

    $query = $this->db->delete(self::TABLE);
    $or = $query->orConditionGroup()
      ->condition(
        // Delete records when the entity acts like a target entity.
        $query->andConditionGroup()
          ->condition('target_type', $entityTypeId)
          ->condition($this->getIdColumnName($entityTypeId, EntityRole::Target), $entityId)
      )
      ->condition(
        // Or when the entity acts like a source entity.
        $query->andConditionGroup()
          ->condition('source_type', $entityTypeId)
          ->condition($this->getIdColumnName($entityTypeId, EntityRole::Source), $entityId)
      );
    $query->condition($or)->execute();
  }

  /**
   * Reacts after an entity revision has been deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $revision
   *   The entity revision being deleted.
   */
  protected function onEntityRevisionDelete(EntityInterface $revision): void {
    $this->getQuery($revision)->condition('source_revision', $revision->getRevisionId())->execute();
  }

  /**
   * Reacts after an entity translation has been deleted.
   *
   * @param \Drupal\Core\Entity\EntityInterface $translation
   *   The entity translation being deleted.
   */
  protected function onEntityTranslationDelete(EntityInterface $translation): void {
    $langcode = $translation->language()->getId();
    $this->getQuery($translation)->condition('source_langcode', $langcode)->execute();
  }

  /**
   * Provides a base query to delete records.
   *
   * @param \Drupal\Core\Entity\EntityInterface $entity
   *   The entity revision or translation.
   *
   * @return \Drupal\Core\Database\Query\Delete
   *   A delete query.
   */
  protected function getQuery(EntityInterface $entity): Delete {
    $entityTypeId = $entity->getEntityTypeId();
    $entityId = $this->getIdColumnValueByEntity($entity);
    return $this->db->delete(self::TABLE)
      ->condition('source_type', $entityTypeId)
      ->condition($this->getIdColumnName($entityTypeId, EntityRole::Source), $entityId);
  }

  /**
   * Returns entity translations.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $entity
   *   The entity.
   *
   * @return iterable<FieldableEntityInterface>
   *   Translated entities.
   */
  protected function getTranslations(FieldableEntityInterface $entity): iterable {
    if ($entity instanceof TranslatableInterface) {
      foreach ($entity->getTranslationLanguages() as $langcode => $language) {
        if ($entity->hasTranslation($langcode)) {
          yield $entity->getTranslation($langcode);
        }
      }
    }
    else {
      yield $entity;
    }
  }

  /**
   * Re-queues source entities if a traversable entity has changed.
   *
   * When a traversable entity is updated or deleted, it might impact the usage
   * of source entities. We search if this traversable entity is part of the
   * usage path and re-queue the source entities where this traversable entity
   * is involved.
   *
   * @param \Drupal\Core\Entity\FieldableEntityInterface $traversableEntity
   *   The affected traversable entity.
   * @param \Drupal\track_usage\Entity\TrackConfigInterface $config
   *   The track config.
   */
  protected function requeueSourceEntities(FieldableEntityInterface $traversableEntity, TrackConfigInterface $config): void {
    $fields = ['config', 'source_type', 'source_id_string', 'source_revision'];
    $rows = $this->db->select(self::TABLE)
      ->distinct()
      ->fields(self::TABLE, $fields)
      ->where("JSON_CONTAINS(meta, JSON_QUOTE(:entity), '\$.paths')", [
        ':entity' => $this->getKey($traversableEntity),
      ])
      ->execute()
      ->fetchAll($this->fetchMode(\PDO::FETCH_NUM));

    foreach ($rows as $row) {
      [$configId, $entityTypeId, $entityIdAsString, $revisionId] = $row;

      $entityIdColumn = $this->getIdColumnName($entityTypeId, EntityRole::Source);
      $entityId = $this->getIdColumnValue($entityTypeId, $entityIdAsString);

      $this->db->delete(self::TABLE)
        ->condition('config', $config->id())
        ->condition('source_type', $entityTypeId)
        ->condition($entityIdColumn, $entityId)
        ->condition('source_revision', intval($revisionId))
        ->execute();

      // Re-queue this source entity.
      $this->recordRaw($entityTypeId, $entityIdAsString, $revisionId, $configId);
    }
  }

}
